#include <msp/core/raii.h>
#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_instanced_arrays.h>
#include <msp/gl/extensions/arb_vertex_array_object.h>
#include <msp/gl/extensions/arb_vertex_attrib_binding.h>
#include <msp/gl/extensions/arb_vertex_buffer_object.h>
#include <msp/gl/extensions/arb_vertex_shader.h>
#include "buffer.h"
#include "gl.h"
#include "vertexarray.h"
#include "vertexsetup.h"

namespace Msp {
namespace GL {

VertexSetup::VertexSetup():
	dirty(0),
	vertex_array(0),
	inst_array(0),
	index_buffer(0)
{
	static Require req(ARB_vertex_array_object);
	if(ARB_direct_state_access)
		glCreateVertexArrays(1, &id);
	else
		glGenVertexArrays(1, &id);
}

VertexSetup::~VertexSetup()
{
	if(current()==this)
		unbind();
	glDeleteVertexArrays(1, &id);
}

void VertexSetup::set_vertex_array(const VertexArray &a)
{
	vertex_array = &a;
	update(VERTEX_ARRAY);
}

void VertexSetup::set_instance_array(const VertexArray *a)
{
	if(a)
		static Require req(ARB_instanced_arrays);

	inst_array = a;
	update(INSTANCE_ARRAY);
}

void VertexSetup::set_index_buffer(const Buffer &ibuf)
{
	index_buffer = &ibuf;
	update(INDEX_BUFFER);
}

void VertexSetup::update(unsigned mask) const
{
	static bool direct = ARB_direct_state_access && ARB_vertex_attrib_binding;
	if(!direct && current()!=this)
	{
		dirty |= mask;
		return;
	}

	if(mask&VERTEX_ARRAY)
		update_vertex_array(*vertex_array, 0, 0, direct);

	if(mask&INSTANCE_ARRAY)
	{
		if(inst_array)
			update_vertex_array(*inst_array, 1, 1, direct);
	}

	if(mask&INDEX_BUFFER)
	{
		if(direct)
			glVertexArrayElementBuffer(id, index_buffer->get_id());
		else
			glBindBuffer(ELEMENT_ARRAY_BUFFER, index_buffer->get_id());
	}
}

void VertexSetup::update_vertex_array(const VertexArray &array, unsigned binding, unsigned divisor, bool direct) const
{
	Conditional<Bind> bind_vbuf(!direct, array.get_buffer(), ARRAY_BUFFER);

	const VertexFormat &fmt = array.get_format();
	unsigned stride = get_stride(fmt)*sizeof(float);
	if(direct)
	{
		glVertexArrayVertexBuffer(id, binding, array.get_buffer()->get_id(), 0, stride);
		glVertexArrayBindingDivisor(id, binding, divisor);
	}

	unsigned offset = 0;
	for(const unsigned char *c=fmt.begin(); c!=fmt.end(); ++c)
	{
		unsigned t = get_component_type(*c);
		if(t>=get_component_type(ATTRIB1))
			t -= get_component_type(ATTRIB1);
		unsigned sz = get_component_size(*c);
		if(direct)
		{
			if(*c==COLOR4_UBYTE)
				glVertexArrayAttribFormat(id, t, 4, GL_UNSIGNED_BYTE, true, offset);
			else
				glVertexArrayAttribFormat(id, t, sz, GL_FLOAT, false, offset);
			glVertexArrayAttribBinding(id, t, binding);
			glEnableVertexArrayAttrib(id, t);
		}
		else
		{
			if(*c==COLOR4_UBYTE)
				glVertexAttribPointer(t, 4, GL_UNSIGNED_BYTE, true, stride, reinterpret_cast<unsigned char *>(offset));
			else
				glVertexAttribPointer(t, sz, GL_FLOAT, false, stride, reinterpret_cast<float *>(offset));
			glVertexAttribDivisor(t, divisor);
			glEnableVertexAttribArray(t);
		}
		offset += sz*sizeof(float);
	}
}

void VertexSetup::bind() const
{
	if(set_current(this))
	{
		glBindVertexArray(id);
		if(dirty)
		{
			update(dirty);
			dirty = 0;
		}
	}
}

void VertexSetup::unbind()
{
	if(set_current(0))
		glBindVertexArray(0);
}

} // namespace GL
} // namespace Msp
